import ducc0
import numpy as np
from time import time

rng = np.random.default_rng(48)


def nalm(lmax, mmax):
    return ((mmax+1)*(mmax+2))//2 + (mmax+1)*(lmax-mmax)


def random_alm(lmax, mmax, spin, ncomp):
    res = rng.uniform(-1., 1., (ncomp, nalm(lmax, mmax))) \
     + 1j*rng.uniform(-1., 1., (ncomp, nalm(lmax, mmax)))
    # make a_lm with m==0 real-valued
    res[:, 0:lmax+1].imag = 0.
    ofs=0
    for s in range(spin):
        res[:, ofs:ofs+spin-s] = 0.
        ofs += lmax+1-s
    return res


def test(lmax, geometry, spin, nthreads=1):
    print("testing lmax={}, spin={}, nthreads={}, geometry={}".format(lmax,spin,nthreads,geometry))
    ncomp = 1 if spin == 0 else 2

    nrings = lmax+1
    if geometry=="CC":
        nrings = lmax+2
    elif geometry=="DH":
        nrings = 2*lmax+2
    elif geometry=="F2":
        nrings = 2*lmax+1
    nphi=2*lmax+2

    alm = random_alm(lmax, lmax, spin, ncomp)
    map = np.empty((alm.shape[0], nrings,nphi))
    alm2 = np.empty((map.shape[0],nalm(lmax,lmax)),dtype=np.complex128)
    t0=time()
    map = ducc0.sht.synthesis_2d(alm=alm, lmax=lmax, spin=spin, map=map, nthreads=nthreads, geometry=geometry)
    alm2 = ducc0.sht.analysis_2d(alm=alm2, map=map, lmax=lmax, spin=spin, geometry=geometry, nthreads=nthreads)
    print(time()-t0)
    print("L2 error after full round-trip:", ducc0.misc.l2error(alm2,alm))
    print("L_inf error after full round-trip:", np.max(np.abs(alm2-alm)))

nthr=8
for l0 in [4095]:
    for geometry in ["CC", "F1", "MW", "MWflip", "GL", "DH", "F2"]:
        for spin in [0,1,2]:
            test(lmax=l0, spin=spin, nthreads=nthr, geometry=geometry)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2022-2023 Max-Planck-Society

import ducc0
import numpy as np
from time import time

try:
    import finufft
    have_finufft = True
except ImportError:
    have_finufft = False

class Bench:
    def __init__(self, shape, npoints):
        self._shape = shape
        ndim = len(shape)
        self._coord = (2*np.pi*np.random.uniform(size=(npoints,ndim)) - np.pi).astype(np.float32)
        # cidx = ((self._coord[:,-1]+np.pi)/(2*np.pi)*1000).astype(np.int64)
        # if ndim > 1:
            # cidx += ((self._coord[:,-2]+np.pi)/(2*np.pi)*1000).astype(np.int64)*1000
        # if ndim > 2:
            # cidx += ((self._coord[:,-3]+np.pi)/(2*np.pi)*1000).astype(np.int64)*1000000
        # idx = np.argsort(cidx)
        # self._coord = self._coord[idx]
        self._points = (np.random.uniform(size=npoints)-0.5
                 + 1j * np.random.uniform(size=npoints)-0.5j).astype(np.complex64)
        self._values = (np.random.uniform(size=shape)-0.5
                 + 1j * np.random.uniform(size=shape)-0.5j).astype(np.complex64)
        eps = 1.0001*ducc0.nufft.bestEpsilon(ndim=ndim, singleprec=False)
        self._res_fiducial_1 = ducc0.nufft.nu2u(
            points=self._points.astype(np.complex128),
            coord=self._coord.astype(np.float64),
            forward=True,
            epsilon=eps,
            nthreads=0,
            verbosity=1,
            out=np.empty(shape, dtype=np.complex128))

        self._res_fiducial_2 = ducc0.nufft.u2nu(
            grid=self._values.astype(np.complex128),
            coord=self._coord.astype(np.float64),
            forward=True,
            epsilon=eps,
            nthreads=0,
            verbosity=1)

    def run(self, epsilon, singleprec, nthreads):
        rdtype = np.float32 if singleprec else np.float64
        dtype = np.complex64 if singleprec else np.complex128

        res={}
        res["shape"] = self._shape
        res["npoints"] = self._coord.shape[0]
        res["epsilon"] = epsilon
        res["nthreads"] = nthreads
        res["singleprec"] = singleprec

        shape = self._shape
        ndim = len(shape)
        npoints = self._coord.shape[0]
        coord = self._coord.astype(rdtype)
        points = self._points.astype(dtype)
        values = self._values.astype(dtype)

        out = np.ones(shape, dtype=dtype)
        plan = ducc0.nufft.plan(nu2u=True, coord=coord, grid_shape=shape, epsilon=epsilon, nthreads=nthreads)
        t0 = time()
        res_ducc = plan.nu2u(points=points, forward=True, verbosity=1, out=out)
        res["ducc_trans_1"] = time()-t0
        res["err_ducc_trans_1"] = ducc0.misc.l2error(res_ducc, self._res_fiducial_1)

        out=np.ones(shape=(npoints,), dtype=dtype)
        plan = ducc0.nufft.plan(nu2u=False, coord=coord, grid_shape=shape, epsilon=epsilon, nthreads=nthreads)
        t0 = time()
        res_ducc = plan.u2nu(grid=values, forward=True, verbosity=1, out=out)
        res["ducc_trans_2"] = time()-t0
        res["err_ducc_trans_2"] = ducc0.misc.l2error(res_ducc, self._res_fiducial_2)

        out = np.ones(shape, dtype=dtype)
        t0 = time()
        res_ducc = ducc0.nufft.nu2u(points=points, coord=coord, forward=True, epsilon=epsilon, nthreads=nthreads, verbosity=1, out=out)
        res["ducc_full_1"] = time()-t0
        res["err_ducc_1"] = ducc0.misc.l2error(res_ducc, self._res_fiducial_1)

        out=np.ones(shape=(npoints,), dtype=dtype)
        t0 = time()
        res_ducc = ducc0.nufft.u2nu(grid=values, coord=coord, forward=True, epsilon=epsilon, nthreads=nthreads, verbosity=1, out=out)
        res["ducc_full_2"] = time()-t0
        res["err_ducc_2"] = ducc0.misc.l2error(res_ducc, self._res_fiducial_2)
        return res

    def run_finufft(self, epsilon, singleprec, nthreads, res):
        import finufft
        rdtype = np.float32 if singleprec else np.float64
        dtype = np.complex64 if singleprec else np.complex128

        shape = self._shape
        ndim = len(shape)
        npoints = self._coord.shape[0]
        coord = self._coord.astype(rdtype)
        coord = tuple(np.ascontiguousarray(coord[:,i]) for i in range(coord.shape[1]))
        points = self._points.astype(dtype)
        values = self._values.astype(dtype)

        # Adding the "fftw=0" argument makes execution somewhat faster,
        # but planning can be painfully slow.
        plan1 = finufft.Plan(1, self._shape, 1, eps=epsilon, isign=-1,
                             dtype="complex64" if singleprec else "complex128",
                             nthreads=nthreads, debug=1, fftw=0)
        plan1.setpts(*coord)
        t0 = time()
        res_finufft = plan1.execute(points)
        res["finufft_full_1"] = time()-t0
        res["err_finufft_1"] = ducc0.misc.l2error(res_finufft, self._res_fiducial_1)
        del plan1

        plan2 = finufft.Plan(2, shape, 1, eps=epsilon, isign=-1,
                            dtype="complex64" if singleprec else "complex128",
                            nthreads=nthreads, debug=1, fftw=0)
        plan2.setpts(*coord)
        t0 = time()
        res_finufft = plan2.execute(values)
        res["finufft_full_2"] = time()-t0
        res["err_finufft_2"] = ducc0.misc.l2error(res_finufft, self._res_fiducial_2)

        return res


def plot(res, fname):
    import matplotlib.pyplot as plt
    fct = 1e9/res[0]["npoints"]
    have_finufft = "finufft_full_1" in res[0]
    tducc1 = fct*np.array([r["ducc_full_1"] for r in res])
    tducc2 = fct*np.array([r["ducc_full_2"] for r in res])
    tducct1 = fct*np.array([r["ducc_trans_1"] for r in res])
    tducct2 = fct*np.array([r["ducc_trans_2"] for r in res])
    educc1 = np.array([r["err_ducc_1"] for r in res])
    educc2 = np.array([r["err_ducc_2"] for r in res])
    educct1 = np.array([r["err_ducc_trans_1"] for r in res])
    educct2 = np.array([r["err_ducc_trans_2"] for r in res])
    if have_finufft:
        tfinufft1 = fct*np.array([r["finufft_full_1"] for r in res])
        tfinufft2 = fct*np.array([r["finufft_full_2"] for r in res])
        efinufft1 = np.array([r["err_finufft_1"] for r in res])
        efinufft2 = np.array([r["err_finufft_2"] for r in res])
    eps = np.array([r["epsilon"] for r in res])
    plt.xscale("log")
    plt.yscale("log")
    plt.plot(educc1,tducc1,label="ducc unplanned, type 1")
    plt.plot(educc2,tducc2,label="ducc unplanned, type 2")
    plt.plot(educct1,tducct1,label="ducc planned, type 1")
    plt.plot(educct2,tducct2,label="ducc planned, type 2")
    if have_finufft:
        plt.plot(efinufft1,tfinufft1,label="finufft planned, type 1")
        plt.plot(efinufft2,tfinufft2,label="finufft planned, type 2")
    plt.title("shape={}, npoints={}, nthreads={}".format(res[0]["shape"], res[0]["npoints"], res[0]["nthreads"]))
    plt.xlabel("real error")
    plt.ylabel("ns per nonuniform point")
    plt.legend()
    plt.savefig(fname, bbox_inches='tight')
    plt.close()


def runbench(shape, npoints, nthreads, fname, singleprec=False):
    res=[]
    mybench = Bench(shape, npoints)
    if singleprec:
        epslist = [[2.5e-7, 4.5e-7, 8.2e-7][len(shape)-1], 1e-5, 1e-4, 1e-3, 1e-2]
    else:
        epslist = [[4e-15, 8e-15, 2e-14][len(shape)-1], 1e-13, 1e-12, 1e-11, 1e-10, 1e-9, 1e-8, 1e-7, 1e-6, 1e-5, 1e-4, 1e-3, 1e-2]
    for eps in epslist:
        tres = mybench.run(eps, singleprec, nthreads)
        if have_finufft:
            res.append(mybench.run_finufft(eps, singleprec, nthreads, tres))
        else:
            res.append(tres)
    plot(res, fname)

singleprec = False
# FINUFFT benchmarks
if True:
    runbench((   1000000,),  10000000, 1, "finufft_1d_serial.png"  , singleprec)
    runbench(( 1000,1000,),  10000000, 1, "finufft_2d_serial.png"  , singleprec)
    runbench((100,100,100),  10000000, 1, "finufft_3d_serial.png"  , singleprec)
    runbench((  10000000,), 100000000, 8, "finufft_1d_parallel.png", singleprec)
    runbench(( 3162,3162,), 100000000, 8, "finufft_2d_parallel.png", singleprec)
    runbench((216,216,216), 100000000, 8, "finufft_3d_parallel.png", singleprec)
# NFFT.jl benchmarks
if True:
    runbench(( 512*512,),  512*512, 1, "bench_1d.png", singleprec)
    runbench(( 512,512,),  512*512, 1, "bench_2d.png", singleprec)
    runbench((64,64,64,), 64*64*64, 1, "bench_3d.png", singleprec)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2020-2022 Max-Planck-Society


import ducc0.healpix as ph
import numpy as np
import math
import pytest
from numpy.testing import assert_equal

pmp = pytest.mark.parametrize


def list2fixture(lst):
    @pytest.fixture(params=lst)
    def myfixture(request):
        return request.param

    return myfixture


pow2 = [1 << shift for shift in range(29)]
nonpow2 = [i+7 for i in pow2]
nside_nest = list2fixture(pow2)
nside_ring = list2fixture(pow2+nonpow2)

vlen = list2fixture([1, 10, 100, 1000, 10000])
ftype = list2fixture([np.float32, np.float64])
itype = list2fixture([np.int32, np.int64])


def random_ptg(rng, vlen):
    res = np.empty((vlen, 2), dtype=np.float64)
    res[:, 0] = np.arccos((rng.random(vlen)-0.5)*2)
#    res[:, 0] = math.pi*rng.random(vlen)
    res[:, 1] = rng.random(vlen)*2*math.pi
    return res


def test_pixangpix_nest(vlen, nside_nest, itype):
    if itype == np.int32 and nside_nest > 8192:
        return
    base = ph.Healpix_Base(nside_nest, "NEST")
    rng = np.random.default_rng(42)
    inp = rng.integers(low=0, high=12*nside_nest*nside_nest-1, size=vlen)
    out = base.ang2pix(base.pix2ang(inp.astype(itype)))
    assert_equal(inp, out)


def test_pixangpix_ring(vlen, nside_ring, itype):
    if itype == np.int32 and nside_ring > 8192:
        return
    base = ph.Healpix_Base(nside_ring, "RING")
    rng = np.random.default_rng(42)
    inp = rng.integers(low=0, high=12*nside_ring*nside_ring-1, size=vlen)
    out = base.ang2pix(base.pix2ang(inp))
    assert_equal(inp, out)


def test_vecpixvec_nest(vlen, nside_nest, ftype):
    if ftype == np.float32 and nside_nest > 8192:
        return
    base = ph.Healpix_Base(nside_nest, "NEST")
    rng = np.random.default_rng(42)
    inp = ph.ang2vec(random_ptg(rng, vlen).astype(ftype)).astype(ftype)
    out = base.pix2vec(base.vec2pix(inp)).astype(ftype)
    assert_equal(np.all(ph.v_angle(inp, out) < base.max_pixrad()), True)


def test_vecpixvec_ring(vlen, nside_ring, ftype):
    if ftype == np.float32 and nside_ring > 8192:
        return
    base = ph.Healpix_Base(nside_ring, "RING")
    rng = np.random.default_rng(42)
    inp = ph.ang2vec(random_ptg(rng, vlen).astype(ftype)).astype(ftype)
    out = base.pix2vec(base.vec2pix(inp)).astype(ftype)
    assert_equal(np.all(ph.v_angle(inp, out) < base.max_pixrad()), True)


def test_ringnestring(vlen, nside_nest, itype):
    if itype == np.int32 and nside_nest > 8192:
        return
    base = ph.Healpix_Base(nside_nest, "NEST")
    rng = np.random.default_rng(42)
    inp = rng.integers(low=0, high=12*nside_nest*nside_nest-1, size=vlen)
    out = base.ring2nest(base.nest2ring(inp.astype(itype)).astype(itype))
    assert_equal(np.all(out == inp), True)


def test_vecangvec(vlen, ftype):
    rng = np.random.default_rng(42)
    inp = random_ptg(rng, vlen).astype(ftype)
    out = ph.vec2ang(ph.ang2vec(inp))
    assert_equal(np.all(np.abs(out-inp) < 1e-14), True)



import sys
import os.path
import itertools
from glob import iglob
import os

from setuptools import setup, Extension
import pybind11

pkgname = 'ducc0'
version = '0.34.0'

tmp = os.getenv("DUCC0_CFLAGS", "").split(" ")
user_cflags = [x for x in tmp if x != ""]
tmp = os.getenv("DUCC0_LFLAGS", "").split(" ")
user_lflags = [x for x in tmp if x != ""]
tmp = os.getenv("DUCC0_FLAGS", "").split(" ")
tmp = [x for x in tmp if x != ""]
user_cflags += tmp
user_lflags += tmp

compilation_strategy = os.getenv('DUCC0_OPTIMIZATION', 'native-strip')
if compilation_strategy not in ['none', 'none-debug', 'none-strip', 'portable', 'portable-debug', 'portable-strip', 'native', 'native-debug', 'native-strip']:
    raise RuntimeError('unknown compilation strategy')
do_debug = compilation_strategy in ['none-debug', 'portable-debug', 'native-debug']
do_strip = compilation_strategy in ['none-strip', 'portable-strip', 'native-strip']
do_optimize = compilation_strategy not in ['none', 'none-debug', 'none-strip']
do_native = compilation_strategy in ['native', 'native-debug', 'native-strip']

def _print_env():
    import platform
    print("")
    print("Build environment:")
    print("Platform:     ", platform.platform())
    print("Machine:      ", platform.machine())
    print("System:       ", platform.system())
    print("Architecture: ", platform.architecture())
    print("")

def _get_files_by_suffix(directory, suffix):
    path = directory
    iterable_sources = (iglob(os.path.join(root, '*.'+suffix))
                        for root, dirs, files in os.walk(path))
    return list(itertools.chain.from_iterable(iterable_sources))


include_dirs = ['.', './src/',
                pybind11.get_include(True),
                pybind11.get_include(False)]

extra_compile_args = ['-std=c++17', '-fvisibility=hidden']

if do_debug:
    extra_compile_args += ['-g']
else:
    extra_compile_args += ['-g0']

if do_optimize:
    extra_compile_args += ['-ffast-math', '-O3']
else:
    extra_compile_args += ['-O0']

if do_native:
    extra_compile_args += ['-march=native']

python_module_link_args = []

define_macros = [("PKGNAME", pkgname),
                 ("PKGVERSION", version)]

if sys.platform == 'darwin':
    extra_compile_args += ['-mmacosx-version-min=10.14', '-pthread']
    python_module_link_args += ['-mmacosx-version-min=10.14', '-pthread']
elif sys.platform == 'win32':
    extra_compile_args = ['/EHsc', '/std:c++17']
    if do_optimize:
        extra_compile_args += ['/Ox']
else:
    if do_optimize:
        extra_compile_args += ['-Wfatal-errors',
                               '-Wfloat-conversion',
                               '-W',
                               '-Wall',
                               '-Wstrict-aliasing',
                               '-Wwrite-strings',
                               '-Wredundant-decls',
                               '-Woverloaded-virtual',
                               '-Wcast-qual',
                               '-Wcast-align',
                               '-Wpointer-arith',
                               '-Wnon-virtual-dtor',
                               '-Wzero-as-null-pointer-constant']
    extra_compile_args += ['-pthread']
    python_module_link_args += ['-Wl,-rpath,$ORIGIN', '-pthread']
    if do_native:
        python_module_link_args += ['-march=native']
    if do_strip:
        python_module_link_args += ['-s']

extra_compile_args += user_cflags
python_module_link_args += user_lflags

depfiles = (_get_files_by_suffix('.', 'h') +
            _get_files_by_suffix('.', 'cc') +
            ['setup.py'])

extensions = [Extension(pkgname,
                        language='c++',
                        sources=['python/ducc.cc'],
                        depends=depfiles,
                        include_dirs=include_dirs,
                        define_macros=define_macros,
                        extra_compile_args=extra_compile_args,
                        extra_link_args=python_module_link_args)]

_print_env()

setup(name=pkgname,
      version=version,
      ext_modules = extensions
      )
